//////////////////////////////////////////////////////////////////
// Symbolic
//
// Library for the construction of nested symbolic propositions
// The Flying Pig!
// Started 5/8/2003
// http://www.flypig.co.uk
// 
// 
// The MIT License
// See symbolic.h, COPYING file or website for licence
// 
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
// Includes

#include "symbolic.h"
#include "symbolic_private.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

//////////////////////////////////////////////////////////////////
// Defines

#define CONTINUED_FRACTION_ERROR (1.0e-10)

//////////////////////////////////////////////////////////////////
// Structures

//////////////////////////////////////////////////////////////////
// Global variables

//////////////////////////////////////////////////////////////////
// Function prototypes

Operation * RecurseToOperation      (char const * szString, int nStrLen);
char      * RecurseToString         (Operation * psOp, int nStrLen);
inline bool StringCheckBinary       (char const * szString, int const nStrLen, char const * szOperator);
bool        TryStringToDouble       (char const * const szString, int const nStrLen, double * pfDecimal);

//////////////////////////////////////////////////////////////////
// Main application

//////////////////////////////////////////////////////////////////
// Turn a formula into a string
char * OperationToString (Operation * psOp, char * szString, int nStrLen) {
  char          *szRecurse;

  szRecurse = RecurseToString (psOp, nStrLen);
  strncpy (szString, szRecurse, nStrLen);
  szString[nStrLen - 1] = 0;
  PropFree (szRecurse);

  return szString;
}

//////////////////////////////////////////////////////////////////
// Recursively turn a formula into a string
char * RecurseToString (Operation * psOp, int nStrLen) {
  char          *szReturn;
  char          *szVar1;
  char          *szVar2;
  char          *szVar3;

  szReturn = (char*)PropCalloc (nStrLen, 1);

  if (psOp) {
    switch (psOp->eOpType) {
      case OPTYPE_MEMORY:
        snprintf (szReturn, nStrLen, "M");
        szReturn[nStrLen - 1] = 0;
        break;
      case OPTYPE_INTEGER:
        snprintf (szReturn, nStrLen, "%d", psOp->Vars.nInteger);
        szReturn[nStrLen - 1] = 0;
        break;
      case OPTYPE_TRUTHVALUE:
        if (psOp->Vars.boTruth) {
          snprintf (szReturn, nStrLen, "TRUE");
          szReturn[nStrLen - 1] = 0;
        }
        else {
          snprintf (szReturn, nStrLen, "FALSE");
          szReturn[nStrLen - 1] = 0;
        }
        break;
      case OPTYPE_VARIABLE:
        snprintf (szReturn, nStrLen, "%s", psOp->Vars.psVar->szVar);
        szReturn[nStrLen - 1] = 0;
        break;
      case OPTYPE_UNARY:
        szVar1 = RecurseToString (psOp->Vars.psUnary->psVar1, nStrLen);
        switch (psOp->Vars.psUnary->eOpType) {
          case OPUNARY_NEGATIVE:
          case OPUNARY_NOT:
            snprintf (szReturn, nStrLen, "%s%s", aszOpUnary[psOp->Vars.psUnary->eOpType], szVar1);
            szReturn[nStrLen - 1] = 0;
            break;
          case OPUNARY_VC:
          case OPUNARY_COSEC:
          case OPUNARY_SINH:
          case OPUNARY_COSH:
          case OPUNARY_TANH:
          case OPUNARY_COS:
          case OPUNARY_SIN:
          case OPUNARY_TAN:
          case OPUNARY_SEC:
          case OPUNARY_COT:
          case OPUNARY_ARCSIN:
          case OPUNARY_ARCCOS:
          case OPUNARY_ARCTAN:
          case OPUNARY_LN:
          case OPUNARY_EXP:
          case OPUNARY_ABS:
            snprintf (szReturn, nStrLen, "%s(%s)", aszOpUnary[psOp->Vars.psUnary->eOpType], szVar1);
            szReturn[nStrLen - 1] = 0;
            break;
          default:
            printf("Invalid unary operator\n");
            break;
        }
        PropFree (szVar1);
        break;
      case OPTYPE_BINARY:
        szVar1 = RecurseToString (psOp->Vars.psBinary->psVar1, nStrLen);
        szVar2 = RecurseToString (psOp->Vars.psBinary->psVar2, nStrLen);
        switch (psOp->Vars.psBinary->eOpType) {
          case OPBINARY_IND:
            snprintf (szReturn, nStrLen, "ind(%s, %s)", szVar1, szVar2);
            szReturn[nStrLen - 1] = 0;
            break;
          case OPBINARY_LAND:
          case OPBINARY_LOR:
          case OPBINARY_LIMP:
          case OPBINARY_POW:
          case OPBINARY_ADD:
          case OPBINARY_SUB:
          case OPBINARY_MUL:
          case OPBINARY_DIV:
          case OPBINARY_DIVIDE:
          case OPBINARY_MOD:
          case OPBINARY_OR:
          case OPBINARY_AND:
          case OPBINARY_BIC:
          case OPBINARY_EOR:
          case OPBINARY_ROR:
          case OPBINARY_ROL:
          case OPBINARY_EQ:
          case OPBINARY_NE:
          case OPBINARY_LT:
          case OPBINARY_LE:
          case OPBINARY_GT:
          case OPBINARY_GE:
            snprintf (szReturn, nStrLen, "(%s %s %s)", szVar1, aszOpBinary[psOp->Vars.psBinary->eOpType], szVar2);
            szReturn[nStrLen - 1] = 0;
            break;
          default:
            printf("Invalid binary operator\n");
            break;
        }
        PropFree (szVar1);
        PropFree (szVar2);
        break;
      case OPTYPE_TERNARY:
        szVar1 = RecurseToString (psOp->Vars.psTernary->psVar1, nStrLen);
        szVar2 = RecurseToString (psOp->Vars.psTernary->psVar2, nStrLen);
        szVar3 = RecurseToString (psOp->Vars.psTernary->psVar3, nStrLen);
        switch (psOp->Vars.psTernary->eOpType) {
          case OPTERNARY_SET:
            snprintf (szReturn, nStrLen, "set(%s, %s, %s)", szVar1,
              szVar2, szVar3);
            szReturn[nStrLen - 1] = 0;
            break;
          default:
            printf("Invalid binary operator\n");
            break;
        }
        PropFree (szVar1);
        PropFree (szVar2);
        PropFree (szVar3);
        break;
      default:
        printf("Invalid operation type\n");
        break;
    }
  }
  else {
    strncpy (szReturn, "", nStrLen);
  }

  return szReturn;
}

//////////////////////////////////////////////////////////////////
// Turn a string into a formula
Operation * StringToOperation (char const * szString) {
  int       nStrLen;
  Operation * psOperation;
  char      * szNoSpaces;
  int       nStrPos;
  int       nNoSpacePos;

  nStrLen = (int)strlen (szString);

  // Remove all spaces
  szNoSpaces = (char *)PropMalloc (nStrLen + 1);
  nNoSpacePos = 0;
  for (nStrPos = 0; nStrPos < nStrLen; nStrPos++) {
    if (szString[nStrPos] != ' ') {
      szNoSpaces[nNoSpacePos] = szString[nStrPos];
      nNoSpacePos++;
    }
  }
  szNoSpaces[nNoSpacePos] = '\0';

  // Now turn it in to an operation recursively
  psOperation = RecurseToOperation (szNoSpaces, nNoSpacePos);

  // Free up our copy of the string with no spaces
  PropFree (szNoSpaces);

  return psOperation;
}

//////////////////////////////////////////////////////////////////
// Check a string fragment to see if it's a binary operator
inline bool StringCheckBinary (char const * szString, int const nStrLen, char const * szOperator) {
  bool      boMatch = FALSE;
  int       nOperatorLen;

  nOperatorLen = (int)strlen (szOperator);

  if ((nStrLen >= nOperatorLen) && (strncmp (szString, szOperator, nOperatorLen) == 0)) {
    boMatch = TRUE;
  }

  return boMatch;
}

//////////////////////////////////////////////////////////////////
// Recursively turn a string into a formula
Operation * RecurseToOperation (char const * szString, int nStrLen) {
  int       nBrackets;
  int       nPos = 0;
  bool      boMatch;
  OPUNARY   eUnary = OPUNARY_INVALID;
  OPBINARY  eBinary = OPBINARY_INVALID;
  Operation * psReturnOp = NULL;
  int       nRightStart;
  double    fDecimal;
  bool      boScanned;
  char      * szVariable = NULL;

  // Remove the edge brackets
  boMatch = TRUE;
  while ((nStrLen > 1) && boMatch && (szString[0] == '(') && (szString[(nStrLen - 1)] == ')')) {
    // Check whether these brackets match
    nBrackets = 0;
    for (nPos = 0; ((nPos < (nStrLen - 1)) && (boMatch)); nPos++) {
      if (szString[nPos] == '(') {
        nBrackets++;
      }
      if (szString[nPos] == ')') {
        nBrackets--;
      }
      if (nBrackets == 0) {
        boMatch = FALSE;
      }
    }
    if (boMatch) {
      szString++;
      nStrLen -= 2;
    }
  }

  // Find the operation with highest precedent
  boMatch = FALSE;
  eBinary = (OPBINARY)((int)OPBINARY_INVALID + 1);
  while ((!boMatch) && (eBinary < OPBINARY_NUM)) {
    nBrackets = 0;
    // Note we start with nPos = 1, so because the LHS of the binary
    // operation has to exist in this case
    for (nPos = 0; ((nPos < nStrLen) && (!boMatch)); nPos++) {
      if (szString[nPos] == '(') {
        nBrackets++;
      }
      if (szString[nPos] == ')') {
        nBrackets--;
      }
      if ((nBrackets == 0) && (nPos > 0)) {
        boMatch = StringCheckBinary (szString + nPos, nStrLen - nPos, aszOpBinary[eBinary]);
      }
    }
    eBinary = (OPBINARY)((int)eBinary + 1);
  }

  if (boMatch) {
    // Split into two pieces and recurse
    eBinary = (OPBINARY)((int)eBinary - 1);
    nRightStart = nPos + (int)strlen(aszOpBinary[eBinary]) - 1;

    psReturnOp = CreateBinary (eBinary, RecurseToOperation (szString, nPos - 1), RecurseToOperation (szString + nRightStart, nStrLen - nRightStart));
  }
  else {
    // Check if it's a unary operation
    boMatch = FALSE;
    eUnary = (OPUNARY)((int)OPUNARY_INVALID + 1);
    while ((!boMatch) && (eUnary < OPUNARY_NUM)) {
      if (nStrLen > (int)strlen (aszOpUnary[eUnary])) {
        if (strncmp (aszOpUnary[eUnary], szString, strlen (aszOpUnary[eUnary])) == 0) {
          boMatch = TRUE;
        }
      }

      eUnary = (OPUNARY)((int)eUnary + 1);
    }

    if (boMatch) {
      // Recurse on whatever is left
      eUnary = (OPUNARY)((int)eUnary - 1);
      nRightStart = (int)strlen(aszOpUnary[eUnary]);
      psReturnOp = CreateUnary (eUnary, RecurseToOperation (szString + nRightStart, nStrLen - nRightStart));
    }
    else {
      // Try to interpret as a number, or failing that a variable
      boScanned = TryStringToDouble (szString, nStrLen, & fDecimal);

      if (boScanned) {
        psReturnOp = ContinuedFraction (fDecimal, CONTFRAC_MAXDEPTH);
      }
      else {
        // Interpret as a variable
        szVariable = (char *)PropMalloc (nStrLen + 1);
        strncpy (szVariable, szString, nStrLen);
        szVariable[nStrLen] = '\0';
        psReturnOp = CreateVariable (szVariable);
        PropFree (szVariable);
        szVariable = NULL;
      }
    }
  }

  return psReturnOp;
}

Operation * ContinuedFraction (double fDecimal, unsigned int uMaxDepth) {
  double fFractional;
  Operation * psReturn;

  fFractional = fDecimal - (int)fDecimal;

  if (((fFractional > -CONTINUED_FRACTION_ERROR) && (fFractional < CONTINUED_FRACTION_ERROR)) || (uMaxDepth == 0)) {
    psReturn = CreateInteger ((int)fDecimal);
  }
  else {
    psReturn = CreateBinary (OPBINARY_DIVIDE, CreateInteger (1), ContinuedFraction (1.0 / fFractional, uMaxDepth - 1));
    if ((int)fDecimal != 0) {
      psReturn = CreateBinary (OPBINARY_ADD, CreateInteger ((int)fDecimal), psReturn);
    }
  }

  return psReturn;
}

bool TryStringToDouble (char const * const szString, int const nStrLen, double * pfDecimal) {
  int nScanned = 0;
  char * szCopied;

  szCopied = (char *)PropMalloc (nStrLen + 1);
  if (szCopied) {
    strncpy (szCopied, szString, nStrLen);
    szCopied[nStrLen] = '\0';

    nScanned = sscanf (szCopied, "%lf", pfDecimal);

    PropFree (szCopied);
  }

  return (nScanned == 1);
}
